function [h, Ex] = NYQdesNyquist (M, NCof, BSpec, GridD)
% The Nyquist filter is expressed as H0(z) H1(z). H1(z) is a filter with
% all of its zeros in the stopband. H0(z) has the passband zeros. First
% H1(z) is designed to give the desired stopband response. Then H0(z)
% is designed to force zeros into the overall impulse response. The
% interaction between H0(z) and H1(z) is handled by iterating the design.
% The stopband weighting for H1(z) is computed from the response of H0(z)
% in the stopband region. In this way, the stopband of the overall Nyquist
% filter has the weighted equiripple behaviour specified by the input
% weighting.

% $Id: NYQdesNyquist.m,v 1.4 2009/07/08 14:21:50 pkabal Exp $

NImax = 10;

% Filter lengths
%   H(z) = H0(z) H1(z)
L0 = floor ((NCof-1)/(2*M));
L1 = (NCof-1)/2 - L0;
N0 = 2*L0 + 1;
N1 = 2*L1 + 1;

% Set up the specifications for H1(z)
FT = 1;
Gc = MMdSpec (FT, N1, BSpec, GridD);
FGrid = acos (Gc.x) / (2*pi);   % Normalized frequency

% Save the weighting for H1(z)
W0 = Gc.W;

% Initialization of H0(z)
h0 = zeros (N0, 1);
h0(L0+1) = 1;

Exx = 0;
Exxx = 0;

for (Iter = 1:NImax)

% Update the weights to include the effect of H0(z)
  Gc.W = abs (NYQ_SymFIR (FGrid, h0)) .* W0;

% Design H1(z)
  [h1, ier, Ex] = MMdesLPFIR (FT, N1, Gc);
  if (ier ~= 0)
    error ('DFiltFIR: Error during filter design');
  end

% Solve for the coefficients of H0(z) which force the Nyquist response
  h0 = NYQsolveZC (h1, M, N0);

% Check for convergence (extremal values unchanged). The extremal values
% lie on a finite grid. Unchanged extremal values do not necessarily imply
% unchanged coefficient values. However, the coefficient values should have
% more or less settled down when the extremal values stop changing. Do one
% more iteration after extremal values have settled down.
  if (any (Ex.x ~= Exx) || any (Ex.x ~= Exxx))
    Exxx = Exx;
    Exx = Ex.x;
  else
    break;
  end
  
end
if (Iter > NImax)
  warning ('DFiltNyquist - Too many iterations');
end

% Generate the Nyquist filter
 h = NYQ_ConvSym (h0, h1);
 
% Set the scaling so that the overall Nyquist filter has a middle sample
% with value 1. This scaling corresponds to a gain of approximately M in
% the passband.
h = NYQ_FixZC (h, M);

return

%-------------
function H = NYQ_SymFIR (f, h)
% Calculate the (zero-phase) frequency response of a symmetric FIR filter.
% The number of coefficients must be odd.

NCof = length (h);
if (mod (NCof, 2) ~= 1)
  error ('DFiltNyquist: Number of coefficients must be odd');
end
if (any (h ~= flipud(h)))
  error ('DFiltNyquist: Filter must be symmetric');
end

% Save the half-filter
nM = (NCof-1)/2;
hh = h(nM+1:end);
Nh = length (hh);

% The response is given by (h[0] is in the middle of the filter),
%                   Nc-1
%   H(f) = h[0] + 2 SUM  h[n] cos(2 pi n f) ,
%                   n=1
% The sum is carried out backwards. Since the end coefficients are usually
% smaller than the middle coefficients, this may give a slightly more
% accurate result.

Nf = length (f);
H = zeros (1, Nf);
n = (Nh-1:-1:1)';
for (k = 1:Nf)
  S = cos (2*pi*f(k)*n)' * hh(n+1);
  H(k) = hh(0+1) + 2 * S;
end

return

%-------------
function h = NYQ_FixZC (h, M)
% Fix an odd length symmetric filter to have a middle coefficient of unity
% and zero crossings every M samples.

NCof = length (h);
nM = (NCof - 1) / 2;    % Middle coefficient, based on zero-based indexing

SF = 1 / h(nM+1);
h = SF * h;

% Check zero crossings
i = nM+M:M:NCof-1;
j = fliplr(NCof-1-i);
I = [j i];
ZCErr = max (abs (h(I+1)));
if (ZCErr > 1e-10)
  warning ('DFiltNyquist - Inaccurate zero crossings (Max Err = %.2g)', ...
           ZCErr);
end

% Force exact zero crossings
h(I+1) = 0;

% Force exact symmetry
h = 0.5 * (h + flipud(h));

return

%----------------------
function z = NYQ_ConvSym (x, y)
% Convolve symmetric responses

% This routine computes the convolution sum
%              Ku
%       z(i) = SUM x(k) y(i-k) ,  Kl=max(0,i-Ny+1), Ku=min(i,Nx-1),
%              k=Kl
% where Nx is the number of terms in x(0),...,x(Nx-1) and Ny is the
% number of terms in y(0),...,y(Ny-1).

% The computation of the convolution sum is separated into two parts
% (plus a middle term). In the second part, the summation index runs
% backwards. If the input responses are symmetric, the computation
% will process the sum terms in the same order for z(i) and z(Nz-1-i).
Nx = length (x);
Ny = length (y);
Nz = Nx + Ny - 1;
NxL = floor(Nx/2) - 1;
NxU = Nx - floor(Nx/2);
z = zeros (Nz, 1);
for (i = 0:Nz-1)
  kst = max (0, i-Ny+1);
  kfn = min (i, Nx-1);

  k = kst:min(kfn, NxL);
  SL = x(k+1)' * y(i-k+1);

  k = kfn:-1:max(kst, NxU);
  SU = x(k+1)' * y(i-k+1);

  if (mod (Nx, 2) ~= 0)
    km = (Nx+1)/2 - 1;
    if (km >= kst && km <= kfn)
      SM = x(km+1) * y(i-km+1);
    else
      SM = 0;
    end
  else
    SM = 0;
  end

  z(i+1) = (SL+SU) + SM;
end

return
