function h0 = NYQsolveZC (h1, M, N0)
% This routine solves for the coefficients of the linear phase FIR
% filter H0(z) which will force the response of the cascade filter
% H(z) = H0(z) H1(z) to be a Nyquist response. A Nyquist response h[n]
% has regular zero crossings every M samples,
%       h[0]   = 1,
%       h[n*M] = 0,   for n ...,-3,-2,-1,+1,+2,+3,...
% h1 - Odd length, symmetric filter coefficients
% M  - Zero crossing interval
% N0 - Number of coefficients in h0

% $Id: NYQsolveZC.m,v 1.1 2009/07/06 18:54:11 pkabal Exp $

N1 = length (h1);
if (mod (N1, 2) ~= 1 || mod (N0, 2) ~= 1)
  error ('NYQsolveZC: Number of coefficients must be odd');
end
if (any (h1 ~= flipud(h1)))
  error ('NYQsolveZC: Filter must be symmetric');
end

% Initialization
NCof = N0 + N1 - 1;
L0 = NYQ_NCofXL0L1 (NCof, M);
if (2*L0+1 ~= N0)
  error ('NYQsolveZC: Incompatible filter lengths');
end

% There are several special cases:
% (1) NCof is of the form
%       NCof = 2*M*m+1
%     In this case the coefficients of h[n] at each end are zero. This
%     filter is really the same as a filter of length 2*M*m-1. Note that
%     L1 is unchanged by this change of length, but L0 decreases by one.
%     We handle that case here by solving for h0[n] with the smaller L0,
%     and then adding a leading and trailing zero to L0.
% (2) h1[n] has leading and trailing zeros. In some cases this can lead
%     to an over-determined set of equations. Note that
%       L0(M-1) <= L1 < LO(M-1) + M.
%     If L1 is equal to the lower limit, then there is only a single
%     coefficient, h1[L0(M-1)], in one of the rows of the equations to be
%     solved. If that coefficient is zero, the system of equations is
%     overdetermined. Here we lop off leading and trailing zeros in h1[n]
%     and recalculate the values for L0 and L1. After solving for the
%     (possibly) shorter h0, we add leading and traiing zeros to it to
%     compensate.

NCofs = NCof;
L0s = L0;
if (mod (NCof-1, 2*M) == 0)
  NCofs = NCofs - 2;
  L0s = NYQ_NCofXL0L1 (NCofs, M);
end

i = find (h1 ~= 0, 1, 'first');
if (isempty (i))
  error ('NYQsolveZC: All input coefficients are zero');
end
NZ1 = i-1;
L0s = NYQ_NCofXL0L1 (NCofs-2*NZ1, M);

% Solve for the coefficients of h0
N0s = 2*L0s + 1;
h0 = NYQ_solve (h1(NZ1+1:end-NZ1), M, N0s);

% Append zeros if necessary
NZ0 = L0 - L0s;
h0 = [zeros(NZ0,1); h0; zeros(NZ0,1)];

return

% ----- -----
function [L0, L1] = NYQ_NCofXL0L1 (NCof, M)
% Consider the factorization H(z) = H0(z) H1(z), where each filter is an
% odd length symmetrical filter. The filter H0(z) forces zero crossings
% into the overall response so that the overall filter is Nyquist.
% L0 - Number of zero crossings on each side of h[n]. The length of h0[n]
%      is 2*L0+1.
% L1 - Number of coefficients on each side of h1[n]
% NCof - Nuber of coefficients in h[n]
% M  - Zero crossing interval

L0 = floor ((NCof-1)/(2*M));
L1 = (NCof-1)/2 - L0;

return

% ----- -----
function g = NYQ_solve (f, M, Ng)
% f  - Odd length, symmetric filter
% M  - Zero crossing interval
% Ng - Number of coefficients in g (odd)

% Given a symmetrical (odd number of coefficients) filter f, find filter g,
% such that
%  h[nM] = SUM g[k] f[nM-k] = 1, for n=0,
%           k               = 0, otherwise.
% Let Ng = 2*N+1,
% [ f[-NM+N] ... f[-NM+1] f[-NM] f[-NM-1] ... f[-NM-N] ] [ g[-N] ]   [  0  ]
% [  ...     ...  ...      ...    ...     ...  ...     ] [  ...  ]   [ ... ]
% [ f[-M+N]  ... f[-M+1]  f[-M]   f[-M-1] ... f[-M-N]  ] [ g[-1] ]   [  0  ]
% [ f[N]     ... f[1]     f[0]    f[-1]   ... f[-N]    ] [ g[0]  ] = [  1  ] 
% [ f[M+N]   ... f[M+1]   f[M]    f[M-1]  ... f[M-N]   ] [ g[1]  ]   [  0  ]
% [  ...     ...  ...      ...    ...     ...  ...     ] [  ...  ]   [ ... ]
% [ f[NM+N]  ... f[NM+1]  f[NM]   f[NM-1] ...  f[NM-N] ] [ g[N]  ]   [  0  ]
%
% We can partition this matrix computation as follows
%   [ FUL fU   FUR ] [gL  ]   [0]
%   [ fL  f[0] fR  ] [g[0]] = [1]
%   [ FBL fB   FBR ] [gB  ]   [0]
% Each corner has an NxN matrix. Then given that f[n] = f[-n], we see that
%   FUL = J FBR J,
%   FUR = J FBL J,
%   fU = J fB,
%   fL = fR J,  (fR and fL are row vectors)
% where J is the square matrix with ones down the NE-SW diagonal. Pre-
% multiplying with J flips columns (flipud); post-multiplying with J
% flips rows (fliplr). With this setup and a symmetrical right-hand side
% vector, it can be shown that
%   gL = J gB.
% These means we can solve for [g[0], gB] and get gL by symmetry.
%
% Using the symmetry relation for gL, the middle row and bottoms rows
% become
%                    [ J gB ]
%   [ fL  f[0] fR  ] [ g[0] ] = [1] => [ f[0] {fL J } + fR   ] [ g[0] ] = [1]
%   [ FBL fB   FBR ] [ gB   ] = [0]    [ fB   {FBL J} + FBR) ] [ gB   ] = [0]
%
% Here we will form a (1+N)x(N+1+N] matrix, and then pick out the slices
%   [ fL  ]  [ f[0] ], [ fR  ]
%   [ FBL ], [ fB   ], [ FBR ].
% These slices will be used to form a (1+N)x(1+N) matrix.

N = (Ng-1) / 2;
Ff = zeros (N+1, 2*N+1);

% Fill up the bottom half matrix, row-by-row
Nf = length (f);
L = (Nf-1) / 2;
foffs = L+1;
for (n = 0:N)
  kst = max (-N, n*M-L);
  kfn = min (N, n*M+L);
  k = kst:kfn;
  Ff(n+1,k+N+1) = f(n*M - k + foffs);
end

% Pick out the slices
FfL = Ff((0:N)+1,(-N:-1)+N+1);
Ff0 = Ff((0:N)+1,0+N+1);
FfR = Ff((0:N)+1,(1:N)+N+1);

% Fill in the coefficient matrix
Fq = [ Ff0, (fliplr(FfL) + FfR) ];   

% Fill in the right-hand side vector
ch = zeros(N+1, 1);
ch(0+1) = 1;

% Solve the equations for the coefficients of H0(z)
gh = Fq \ ch;

% Fill in the rest of g
gb = gh(1+1:end);
g = [flipud(gb); gh];

return
